@@ -4,7 +4,7 @@ use strict;
 use vars qw($VERSION @ISA);
 use Java::JVM::Classfile;
 
-$VERSION = '0.13';
+$VERSION = '0.15';
 
 sub new {
   my $class = shift;
@@ -51,6 +51,19 @@ sub out {
   return java::io::PrintStream->new();
 }
 
+package java::lang::String;
+sub new {
+  my $class = shift;
+  my $self = {};
+  $self->{value} = "";
+  return bless $self, $class;
+}
+
+sub valueOf {
+  my $class = shift;
+  return $_[0];
+}
+
 package java::lang::StringBuffer;
 sub new {
   my $class = shift;
@@ -70,17 +83,21 @@ sub toString {
 }
 |;
 
-  $code .= "package " . $c->class . ";\n";
+  $code .= "\npackage " . $c->class . ";\n";
+
+  $code .= "no warnings 'recursion';\n";
 
   die "Subclasses not supported!" if $c->superclass ne "java/lang/Object";
 
   foreach my $method (@{$c->methods}) {
     next if $method->name eq '<init>';
-    $code .= "sub " . $method->name . " {\n";
+    $code .= "\nsub " . $method->name . " {\n";
+
     $code .= "my \@stack;\n";
     $code .= "my \$class = shift();\n";
     $code .= "my \@locals = \@_;\n";
-    $code .= "my(\$o, \$p, \$return);\n";
+    $code .= "my(\$o, \$p, \$return, \@in);\n";
+    $code .= "my \@params;\n";
 #    $code .= qq|print "locals ";\n|;
 #    $code .= qq|print join("# ", \@\$locals[0]) . "\\n";\n|;
     foreach my $att (@{$method->attributes}) {
@@ -109,54 +126,40 @@ sub toString {
 	  $class =~ s|/|::|g;
 	  my $method = $args[1];
 	  my $signature = $args[2];
-	  my($in, $out) = $signature =~ /^\((.*?)\)(.*?)$/;
-	  $out = "" if defined($out) && $out eq 'V';
-#	  $code .= "push \@stack, (pop \@stack)->$method(pop \@stack);\n";
-	  if ($in) {
-	    $code .= qq|\$o = pop \@stack;
-\$p = pop \@stack;
-\$return = \$p->$method(\$o); # $in / $out\n|;
-	  } else {
-	    $code .= "\$return = (pop \@stack)->$method(); # $in / $out\n";
-	  }
-	  $code .= "push \@stack, \$return;\n" if $out;
+          $code .= $self->invokevirtual_code($class, $method, $signature);
 	} elsif ($op eq 'invokestatic') {
 	  my $class = $args[0];
 	  $class =~ s|/|::|g;
 	  my $method = $args[1];
 	  my $signature = $args[2];
 	  my($in, $out) = $signature =~ /^\((.*?)\)(.*?)$/;
-	  $out = "" if defined($out) && $out eq 'V';
-#	  $code .= "push \@stack, (pop \@stack)->$method(pop \@stack);\n";
-	  if ($in) {
-	    $code .= qq|\$o = pop \@stack;
-\$return = $class->$method(\$o); # $in / $out\n|;
-	  } else {
-	    $code .= "\$return = $class->$method(); # $in / $out\n";
-	  }
-	  $code .= "push \@stack, \$return;\n" if $out;
+          $code .= $self->invokestatic_code($class, $method, $signature);
 	} elsif ($op eq 'invokespecial') {
 	  $code .= "pop \@stack;\n";
 	} elsif ($op eq 'ldc') {
 	  my $arg = $args[0];
 	  $code .= "push \@stack, '$arg';\n";
+	} elsif ($op eq 'ldc2_w') {
+          my $arg = $args[1];
+          my @cpool = @{$c->constant_pool};
+	  $code .= "push \@stack, ".$cpool[$arg]->values->[0].";\n";
 	} elsif ($op eq 'bipush' or $op eq 'sipush') {
 	  my $arg = $args[0];
 	  $code .= "push \@stack, $arg;\n";
 	} elsif ($op eq 'return') {
 	  $code .= "return;\n";
-	} elsif ($op eq 'ireturn') {
-	  $code .= "return pop(\@stack);\n";
-	} elsif ($op =~ /iconst_(\d)/) {
+      	} elsif ($op =~ /^[id]return$/) {
+          $code .= "return pop(\@stack);\n";
+	} elsif ($op =~ /^iconst_(\d)/) {
 	  $code .= "push \@stack, $1;\n";
-	} elsif ($op =~ /istore_(\d)/) {
-	  $code .= "\$locals[$1] = pop \@stack;\n";
-	} elsif ($op eq 'istore') {
+        } elsif ($op =~ /^[dai]store_(\d)/) {
+          $code .= "\$locals[$1] = pop \@stack;\n";
+	} elsif ($op =~ /^[dai]store/) {
 	  my $i = $args[0];
 	  $code .= "\$locals[$i] = pop \@stack;\n";
-	} elsif ($op =~ /[ai]load_(\d)/) {
+	} elsif ($op =~ /[dai]load_(\d)/) {
 	  $code .= "push \@stack, \$locals[$1];\n";
-	} elsif ($op =~ /^[ai]load$/) {
+	} elsif ($op =~ /^[dai]load$/) {
 	  my $i = $args[0];
 	  $code .= "push \@stack, \$locals[$i];\n";
 	} elsif ($op eq 'goto') {
@@ -164,14 +167,15 @@ sub toString {
 	  $code .= "goto $label;\n";
 	} elsif ($op eq 'dup') {
 	  $code .= "push \@stack, \$stack[-1];\n";
-	} elsif ($op eq 'iadd') {
+	} elsif ($op =~ /^[di]add$/) {
 	  $code .= "push \@stack, (pop \@stack) + (pop \@stack);\n";
-	} elsif ($op eq 'isub') {
+	} elsif ($op =~ /^[di]sub/) {
 	  $code .= "push \@stack, - (pop \@stack) + (pop \@stack);\n";
+	} elsif ($op =~ /^[di]mul/) {
+          $code .= "push \@stack, (pop \@stack) * (pop \@stack);\n";
 	} elsif ($op eq 'aaload') {
 	  $code .= qq|\$o = pop \@stack;
 my \$array = pop \@stack;
-#print join("; ", \@\$array) . "\\n";
 push \@stack, \$array->[\$o];\n|;
 	} elsif ($op eq 'iinc') {
 	  my $i = $args[0];
@@ -185,16 +189,67 @@ push \@stack, \$array->[\$o];\n|;
 	} elsif ($op eq 'if_icmpge') {
 	  my $label = $args[0];
 	  $code .= "goto $label if (pop \@stack) <= (pop \@stack);\n";
+	} elsif ($op eq 'ifne') {
+	  my $label = $args[0];
+	  $code .= "goto $label if (pop \@stack);\n";
 	} else {
 	  $code .= "# ?\n";
 	}
       }
-      print "\n";
     }
-    $code .= "}\n";
+    $code .= "}\n\n";
   }
 #  $code .= qq|print join(", ", \@ARGV) . "\\n";\n|;
   $code .= $c->class . "->main([\@ARGV]);\n";
   return $code;
 }
 
+# Invoking static methods
+sub invokestatic_code {
+    my $self = shift;
+    my ($class, $method, $signature) = @_;
+
+    my ($code, $incount);
+    my($in, $out) = $signature =~ /^\((.*?)\)(.*?)$/;
+
+    $in =~ s/L[^;]*;/L/g;
+    $incount = () = $in =~ /[DIL]/g;
+    $out = "" if defined($out) && $out eq 'V';
+    if ($in) {
+	$code .= qq|
+\@params = splice(\@stack,-$incount);
+\$return = $class->$method(\@params); # $in / $out\n
+|;
+    } else {
+	$code .= "\$return = $class->$method(); # $in / $out\n";
+    }
+    $code .= "push \@stack, \$return;\n" if $out;
+ 
+    return $code;
+}
+
+# Invoking virtual methods
+sub invokevirtual_code {
+    my $self = shift;
+    my ($class, $method, $signature) = @_;
+
+    my ($code, $incount);
+    my($in, $out) = $signature =~ /^\((.*?)\)(.*?)$/;
+    $in =~ s/L[^;]*;/L/g;
+    $incount = () = $in =~ /[DIL]/g;
+    $out = "" if defined($out) && $out eq 'V';
+
+    if ($in) {
+	$code .= qq|
+\@params = splice(\@stack,-$incount);
+\$p = pop \@stack;
+\$return = \$p->$method(\@params); # $in / $out\n|;
+    } else {
+	$code .= "\$return = (pop \@stack)->$method(); # $in / $out\n";
+    }
+    $code .= "push \@stack, \$return;\n" if $out;
+
+    return $code;
+}
+
+1;
